/*
 * V1.4 - 08 Jan 2000
 *
 * Pine v4 SSL add-on hack
 *
 * Copyright 1999 Nick Sayer, All Rights Reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions of either source or binary form must not
 *    violate US export laws.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <openssl/ssl.h>

static SSL_CTX *ctx;

struct ssl_stream {
   TCPSTREAM *tcp_stream;
   SSL *ssl;
   char *read_buf;
   int read_len;
};

typedef struct ssl_stream SSL_STREAM;

static long ssl_getbuffer();
static long ssl_readmore();
static void ssl_clear();

/* MAHOGANY: boolean to check if we ought to run */
static int ssl_initialised = 0;

/*
 * Open a normal connection to the host/service/port triple.
 * Mostly, those 3 params will be just passed to tcp_open().
 * Note that the 'service' has already had the ALT_*NAME
 * substitution, which means we will be passing tcp_open()
 * a name starting with '*'. This will make tcp_open() not
 * do a bunch of read-ahead stuff we don't want it to do.
 *
 * Return either a void* that will be passed on future calls
 * or NIL for failure.
 */
static void *ssl_open(host, service, port)
   char *host,*service;
   unsigned long port;
{
   SSL_STREAM *ssl;

   /* MAHOGANY: do we have SSL ? */
   if( ! ssl_initialised )
   {
      mm_log("SSL library not initialised, should never happen!",ERROR);
      return NIL;
   }
   ssl=(SSL_STREAM *)malloc(sizeof(SSL_STREAM));

   if ((ssl->ssl=SSL_new(ctx))==NULL)
   {
      mm_log("SSL library error, should never happen!",ERROR);
      free(ssl);
      return NIL;
   }
   if ((ssl->tcp_stream=tcp_open(host,service,port))==NULL)
   {
      SSL_free(ssl->ssl);
      free(ssl);
      return NIL; /* tcp_open will mm_log() */
   }

   SSL_set_rfd(ssl->ssl,ssl->tcp_stream->tcpsi);
   SSL_set_wfd(ssl->ssl,ssl->tcp_stream->tcpso);
   SSL_set_read_ahead(ssl->ssl,1);
   if (SSL_connect(ssl->ssl)<0)
   {
      char string1[120],string2[150];

      ERR_error_string(ERR_get_error(), string1);
      sprintf(string2,"SSL protocol error: %s",string1);
      mm_log(string2,ERROR);
      tcp_close(ssl->tcp_stream);
      SSL_free(ssl->ssl);
      free(ssl);
      return NIL;
   }

   {
      char string[120];
      int b;

      SSL_get_cipher_bits(ssl->ssl,&b);
      sprintf(string,"SSL connection type: %s (%d bits)",
              SSL_get_cipher(ssl->ssl),b);
      mm_log(string,WARN);
   }

   /*
    * At this point, we really should analyze and/or display
    * the server certificate.
    */

   ssl->read_buf=NULL;
   ssl->read_len=0;

   return ssl;
}

/*
 * I don't even know what this is for. In unix_tcp.c, this section
 * of code appears to do the rsh-to-imap style of open.
 * Returning NIL always appears to work ok.
 */
static void *ssl_aopen(mb, service, usrbuf)
   NETMBX *mb;
   char *service,*usrbuf;
{
   return NIL;
}

/*
 * Get a line of input from the given stream.
 * Return it in a freshly malloc()ed buffer.
 * Presume line termination is two characters long,
 * starting with either \015 or \012. Eat the second
 * character, and replace the first with \0 (C style
 * end of line).
 *
 * Input should not block. If you can't read before
 * you get a CR or LF, return NIL and store what you
 * read so far so that perhaps you can finish it on the
 * next call.
 */
static char *ssl_getline(stream)
   SSL_STREAM *stream;
{
   int n;
   char *outline;

   /* MAHOGANY: do we have SSL ? */
   if( ! ssl_initialised )
   {
      mm_log("SSL library not initialised, should never happen!",ERROR);
      return NIL;
   }   
   for(n=0;;n++) {
      if (n>=stream->read_len)
         if (!ssl_readmore(stream))
            return NIL;

      if (*(stream->read_buf+n)=='\n' || *(stream->read_buf+n)=='\r')
         break;
   }

   outline=malloc(n+3);
   while(ssl_getbuffer(stream,n+2,outline)==NIL) mm_log("getbufr NIL!",NIL);
   outline[n]='\0';
   outline=realloc(outline,n+1);
   if (outline==NULL)
      mm_log("realloc ate output line! Should never happen!",ERROR);

   return outline;
}

/*
 * Fill a buffer from the given stream. Note that the size is actually
 * one more than the read size, since this routine is supposed to \0
 * terminate the buffer (believe it or not).
 *
 * This routine should not block. Return T if you satisfy the read,
 * otherwise return NIL and buffer the stuff you read so far in the
 * hope that the next try will work.
 *
 */
static long ssl_getbuffer(stream, size, buffer)
   SSL_STREAM *stream;
   unsigned long size;
   char *buffer;
{
   /* MAHOGANY: do we have SSL ? */
   if( ! ssl_initialised )
   {
      mm_log("SSL library not initialised, should never happen!",ERROR);
      return NIL;
   }   

   while(size>stream->read_len && ssl_readmore(stream)) ;

   if (size>stream->read_len)
      return NIL;

   bcopy(stream->read_buf,buffer,size);

   ssl_clear(stream,size);

   buffer[size]='\0'; /* THAT'S friggin' goofey! Check tcp_unix.c if you
                         don't believe me */
   return T;
}

/*
 * This is an internal routine. It extends the readahead buffer.
 * It takes an SSL structure pointer, and if there's any data
 * pending, lengthens the buffer by that much and reads it in.
 * Returns T if it did any work, NIL otherwise.
 */

static long ssl_readmore(ssl)
   SSL_STREAM *ssl;
{
   int n;

   /* MAHOGANY: do we have SSL ? */
   if( ! ssl_initialised )
   {
      mm_log("SSL library not initialised, should never happen!",ERROR);
      return NIL;
   }   
   {
      char toss; /* "kick start" SSL_pending */
      SSL_read(ssl->ssl,&toss,0);
   }

   if ((n=SSL_pending(ssl->ssl))<=0) {
      mm_log("ssl_readmore returning NIL",NIL);
      return NIL;
   }

   if ((ssl->read_buf=realloc(ssl->read_buf,ssl->read_len+n))==NULL) {
      mm_log("SSL memory allocation error!",ERROR);
      return NIL;
   }

   if (SSL_read(ssl->ssl,ssl->read_buf+ssl->read_len,n)!=n) {
      mm_log("SSL short read, should never happen !",ERROR);
      return NIL;
   }

   ssl->read_len+=n;

   return T;
}

/*
 * n bytes have been read from the buffer. Close ranks.
 */
static void ssl_clear(stream,n)
   SSL_STREAM *stream;
   int n;
{
   int i;

   /* MAHOGANY: do we have SSL ? */
   if( ! ssl_initialised )
      return;
   
   if (n>stream->read_len) {
      mm_log("clearing more than available! Should never happen!",ERROR);
      return ;
   }

   if (n==stream->read_len) {
      free(stream->read_buf);
      stream->read_buf=NULL;
      stream->read_len=0;
      return;
   }

   bcopy(stream->read_buf+n,stream->read_buf,stream->read_len-n);

   stream->read_buf=realloc(stream->read_buf,stream->read_len-n);
   if (stream->read_buf == NULL) {
      mm_log("memory shrink failed!",ERROR);
      stream->read_len=0;
      return;
   }
   stream->read_len-=n;
}

/*
 * send the sized string to stdout. Return T on success, NIL on
 * failure. I suppose this is supposed to be non-blocking, but
 * I don't really care. Pine doesn't do a lot of talking.
 */
static long ssl_sout(stream, string, size)
   SSL_STREAM *stream;
   char *string;
   unsigned long size;
{
   /* MAHOGANY: do we have SSL ? */
   if( ! ssl_initialised )
      return 0;
   
   return (size==SSL_write(stream->ssl,string,size));
}

/*
 * Same as above, but figure out the length with strlen()
 */
static long ssl_soutr(stream, string)
   SSL_STREAM *stream;
   char *string;
{
   /* MAHOGANY: do we have SSL ? */
   if( ! ssl_initialised )
      return 0;
   return ssl_sout(stream,string,strlen(string));
}

/*
 * close the named stream
 */
static void ssl_close(stream)
   SSL_STREAM *stream;
{
   /* MAHOGANY: do we have SSL ? */
   if( ! ssl_initialised )
   {
      mm_log("SSL library not initialised, should never happen!",ERROR);
      return;
   }   
   tcp_close(stream->tcp_stream);
   SSL_free(stream->ssl);
   if (stream->read_buf!=NULL)
      free(stream->read_buf);
   free(stream);
}

/*
 * The rest of these just use the underlying tcp_driver services,
 * so I don't feel like documenting them.
 */
static char *ssl_localhost(stream)
   SSL_STREAM *stream;
{
   return tcp_localhost(stream->tcp_stream);
}

static char *ssl_remotehost(stream)
   SSL_STREAM *stream;
{
   return tcp_remotehost(stream->tcp_stream);
}

static unsigned long ssl_port(stream)
   SSL_STREAM *stream;
{
   return tcp_port(stream->tcp_stream);
}

static char *ssl_host(stream)
   SSL_STREAM *stream;
{
   return tcp_host(stream->tcp_stream);
}

/*
 * This is our real interface with Pine.
 */
static NETDRIVER ssldriver = {
   ssl_open,
   ssl_aopen,
   ssl_getline,
   ssl_getbuffer,
   ssl_soutr,
   ssl_sout,
   ssl_close,
   ssl_host,
   ssl_remotehost,
   ssl_port,
   ssl_localhost,
};

/*
 * This is our dummy initialization routine. See below for how this
 * ends up getting called. This driver must make a series of calls
 * to inform the c-client layer that an alternate net driver exists.
 * Users can select the alternate driver by matching the parameter
 * to SET_ALTDRIVERNAME after a "/" in a remote server spec
 * (e.g. {localhost/imap/ssl}). When it is selected, the standard
 * service names (imap, pop3, smtp and nntp) will be substituted
 * for the alternate ones listed. The '*' will be noted by tcp_open()
 * and stripped off. Also, the port numbers (used if either the user
 * specifies it or getservbyname() fails) will be substituted as well.
 */

/* MAHOGANY: we don't do anything here, but call the
   ssl_init_Mahogany() from within Mahogany once we loaded the SSL
   libraries, and only if they are available.
*/
static long ssl_init()
{
   return NIL;
}

extern long ssl_init_Mahogany()
{
   mail_parameters(NIL,SET_ALTDRIVER,&ssldriver);
   mail_parameters(NIL,SET_ALTDRIVERNAME,"ssl");

   /*
    * We have to tack a '*' on the front of the service. This will
    * disable some adverse look-ahead behavior in tcp_open.
    * We have to do this because we won't be using any of the
    * other functionality of the TCP netdriver, as SSL deals with
    * the socket directly.
    */

   mail_parameters(NIL,SET_ALTIMAPNAME,"*imaps");
   mail_parameters(NIL,SET_ALTPOPNAME,"*pop3s");
   mail_parameters(NIL,SET_ALTNNTPNAME,"*nntps");
   mail_parameters(NIL,SET_ALTSMTPNAME,"*smtps");

   mail_parameters(NIL,SET_ALTIMAPPORT,(void *) 993);
   mail_parameters(NIL,SET_ALTPOPPORT,(void *) 995);
   mail_parameters(NIL,SET_ALTNNTPPORT,(void *) 563);
   mail_parameters(NIL,SET_ALTSMTPPORT,(void *) 465);

   SSL_library_init();
   SSL_load_error_strings();

   /*
    * A few SSL/TLS servers refuse to accept any other method but
    * their own (notably SIMS 3.5 which only likes SSLv3) for connections.
    *
    * According to the now-obsolete SSLv3 specification, SSLv3 servers
    * should accept SSLv2 client hellos, but this is not always the case
    * since RFC 2246 (TLS) calls for SSLv2's demise. RFC 2246 also indicates
    * that TLSv1 servers should accept SSLv3 client hellos since the two
    * protocols are similar.
    *
    * By specifying EXTRACFLAGS=-DSSLV3ONLYSERVER at the end of the
    * usual build command, the client will speak only SSLv3 by default,
    * thus alleviating some difficulties where TLSv1/SSLv2 are not supported.
    * If EXTRACFLAGS=-DTLSV1ONLYSERVER is specified, the client will
    * try to connect using TLSv1. If neither is specified, it will speak
    * SSLv3 and can rollback to SSLv2 (using SSLv23_client_method()).
    *
    * See <openssl/ssl.h> for further information.
    *
    */

#if defined(SSLV3ONLYSERVER) && !defined(TLSV1ONLYSERVER)
   ctx=SSL_CTX_new(SSLv3_client_method());
#elif defined(TLSV1ONLYSERVER) && !defined(SSLV3ONLYSERVER)
   ctx=SSL_CTX_new(TLSv1_client_method());
#else
   ctx=SSL_CTX_new(SSLv23_client_method());
#endif

   /* MAHOGANY: this is our safety check */
   ssl_initialised = 1;

   return NIL; /* Must always return NIL or Pine will think the
                  AUTHENTICATOR struct actually means something */
};

/*
 * We really aren't an authenticator. But it's the most straightforward
 * way to hook into pine. We need to run ssl_init() at startup.
 * By calling ourselves an EXTRAAUTHENTICATORS, we get a line in
 * linkage.c to link this structure into the list of authenticators.
 * In doing so, it calls the valid() function to see whether or not
 * this authenticator works on this machine. The answer will always be
 * "no", but we can hook in our own way and Pine will be n'er the
 * wiser.
 */

AUTHENTICATOR auth_ssl={
   T,
   "ignore",
   ssl_init,
   NULL,
   NULL,
   NULL,
};
